@@ -17,13 +17,21 @@
 # by typing 'glycogenius'. If not, see <https://www.gnu.org/licenses/>.
 
 from . import General_Functions
+from . import Execution_Functions
 from .Execution_Functions import print_sep
 import pkg_resources
 import pathlib
+import shutil
+import tempfile
+import importlib
 import datetime
 import platform
 import os
 
+date = datetime.datetime.now()
+begin_time = str(date)[2:4]+str(date)[5:7]+str(date)[8:10]+"_"+str(date)[11:13]+str(date)[14:16]+str(date)[17:19]
+temp_folder = os.path.join(tempfile.gettempdir(), "gg_"+begin_time)
+
 version1 = "0.0.0"
 version2 = "0.0.0"
 try:
@@ -42,7 +50,7 @@ if version2 > version1:
     version = version2
 else:
     version = version1
-
+    
 def print_header(complete = True):
     '''Prints a default header to be used in CLI.
     
@@ -160,7 +168,10 @@ def interactive_terminal():
     input_order = [None]
     curr_os = platform.system()
     if curr_os == "Linux":
-        home = "/home/"+str(os.path.expanduser("~" if "SUDO_USER" not in os.environ else os.environ["SUDO_USER"]))
+        if "SUDO_USER" not in os.environ:
+            home = str(os.path.expanduser("~"))
+        else:
+            home = "/home/"+str(os.path.expanduser(os.environ["SUDO_USER"]))
         default_path = home+"/GlycoGenius/"
     if curr_os == "Windows":
         default_path = "C:/GlycoGenius/"
@@ -186,9 +197,11 @@ def interactive_terminal():
             print("PROGRAM IS WITH YOU. SHOULD THE PROGRAM PROVE")
             print("DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY")
             print("SERVICING, REPAIR OR CORRECTION.\n")
+            input("\nPress Enter to continue.")
             continue
         if var == 'version':
             print("\nGlycoGenius version: "+version)
+            input("Press Enter to continue.")
             continue
         if var == 'license':
             license_path = str(pathlib.Path(__file__).parent.parent.resolve())
@@ -198,16 +211,17 @@ def interactive_terminal():
             with open(license_path+"/LICENSE.py", 'r') as f:
                 for line in f:
                     print(line, end = "")
+                input("\nPress Enter to continue.")
             continue
         try:
             var = int(var)
         except:
-            print("Wrong Input")
+            input("\nWrong Input. Press Enter to try again.\n")
             continue
         if var == 5:
             os._exit(1)
         if var < 1 or var > 4:
-            print("Wrong Input")
+            input("\nWrong Input. Press Enter to try again.\n")
             continue
         if var > 0 and var <= 4:
             input_order[0] = var
@@ -244,11 +258,11 @@ def interactive_terminal():
                 try:
                     comp = General_Functions.form_to_comp(var)
                 except:
-                    print('Wrong input')
+                    input("\nWrong Input. Press Enter to try again.\n")
                     continue
                 for i in comp:
                     if i != 'H' and i != 'N' and i != 'S' and i != 'F' and i != 'G':
-                        print('Wrong input')
+                        input("\nWrong Input. Press Enter to try again.\n")
                         continue
                 glycans_list.append(var)
                 print("Current glycans: ", glycans_list)
@@ -267,7 +281,7 @@ def interactive_terminal():
                 else:
                     continue
         if input_order[1] == 2:
-            lib_settings = [None, None, None, None, None, None, None, None, None, None, None, None, None, None, None]
+            lib_settings = [None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None]
             for i in range(len(lib_settings)):
                 if i == 0:
                     while lib_settings[0] == None:
@@ -330,6 +344,26 @@ def interactive_terminal():
                             continue
                         lib_settings[5] = var
                 if i == 6:
+                    while lib_settings[4] == None:
+                        var = input("Type the minimum amount of Xyloses (default: 0): ")
+                        if var == '':
+                            lib_settings[15] = 0
+                        try:
+                            var = int(var)
+                        except:
+                            continue
+                        lib_settings[4] = var
+                if i == 7:
+                    while lib_settings[5] == None:
+                        var = input("Type the maximum amount of Xyloses (default: 0): ")
+                        if var == '':
+                            lib_settings[16] = 0
+                        try:
+                            var = int(var)
+                        except:
+                            continue
+                        lib_settings[5] = var
+                if i == 8:
                     while lib_settings[6] == None:
                         var = input("Type the minimum amount of dHex (default: 0): ")
                         if var == '':
@@ -339,7 +373,7 @@ def interactive_terminal():
                         except:
                             continue
                         lib_settings[6] = var
-                if i == 7:
+                if i == 9:
                     while lib_settings[7] == None:
                         var = input("Type the maximum amount of dHex (default: 2): ")
                         if var == '':
@@ -349,7 +383,7 @@ def interactive_terminal():
                         except:
                             continue
                         lib_settings[7] = var
-                if i == 8:
+                if i == 10:
                     while lib_settings[8] == None:
                         var = input("Type the minimum amount of Neu5Ac (default: 0): ")
                         if var == '':
@@ -359,7 +393,7 @@ def interactive_terminal():
                         except:
                             continue
                         lib_settings[8] = var
-                if i == 9:
+                if i == 11:
                     while lib_settings[9] == None:
                         var = input("Type the maximum amount of Neu5Ac (default: 4): ")
                         if var == '':
@@ -369,7 +403,7 @@ def interactive_terminal():
                         except:
                             continue
                         lib_settings[9] = var
-                if i == 10:
+                if i == 12:
                     while lib_settings[10] == None:
                         var = input("Type the minimum amount of Neu5Gc (default: 0): ")
                         if var == '':
@@ -379,7 +413,7 @@ def interactive_terminal():
                         except:
                             continue
                         lib_settings[10] = var
-                if i == 11:
+                if i == 13:
                     while lib_settings[11] == None:
                         var = input("Type the maximum amount of Neu5Gc (default: 0): ")
                         if var == '':
@@ -389,7 +423,7 @@ def interactive_terminal():
                         except:
                             continue
                         lib_settings[11] = var
-                if i == 12:
+                if i == 14:
                     while lib_settings[12] == None:
                         var = input("Type the minimum amount of total sialic acids\n(default: 0): ")
                         if var == '':
@@ -399,7 +433,7 @@ def interactive_terminal():
                         except:
                             continue
                         lib_settings[12] = var
-                if i == 13:
+                if i == 15:
                     while lib_settings[13] == None:
                         var = input("Type the maximum amount of total sialic acids\n(default: 4): ")
                         if var == '':
@@ -409,7 +443,7 @@ def interactive_terminal():
                         except:
                             continue
                         lib_settings[13] = var
-                if i == 14:
+                if i == 16:
                     while lib_settings[14] == None:
                         var = input("Force compositions to N-glycans structure\n (default: yes) (y/n): ")
                         if var == '':
@@ -439,7 +473,7 @@ def interactive_terminal():
             try:
                 var2 = int(var2)
             except:
-                print('Wrong input')
+                input("\nWrong Input. Press Enter to try again.\n")
                 continue
             adducts[var] = var2
         print_sep()
@@ -451,7 +485,7 @@ def interactive_terminal():
             try:
                 var = int(var)
             except:
-                print('Wrong input')
+                input("\nWrong Input. Press Enter to try again.\n")
                 continue
             max_charges = var
             break
@@ -462,7 +496,7 @@ def interactive_terminal():
             if var == 'y' or var == 'n':
                 break
             else:
-                print("Wrong Input")
+                input("\nWrong Input. Press Enter to try again.\n")
                 continue
         print("")
         if var == 'y':
@@ -479,14 +513,14 @@ def interactive_terminal():
                 break
         lacto_eesterified = False
         while True:
-            var = input("Is the sample lactonized/ethyl-esterified? (y/n): ")
+            var = input("Is the sample aminated/ethyl-esterified? (y/n): ")
             if var == 'y' or var == '':
                 lacto_eesterified = True
                 break
             elif var == 'n':
                 break
             else:
-                print("Wrong Input")
+                input("\nWrong Input. Press Enter to try again.\n")
                 continue
         permethylated = False
         reduced = False
@@ -499,7 +533,7 @@ def interactive_terminal():
                 if var == 'n':
                     break
                 else:
-                    print('Wrong input')
+                    input("\nWrong Input. Press Enter to try again.\n")
                     continue
             while True:
                 var = input("Are the glycans reduced (y/n): ")
@@ -509,7 +543,7 @@ def interactive_terminal():
                 if var == 'n':
                     break
                 else:
-                    print('Wrong input')
+                    input("\nWrong Input. Press Enter to try again.\n")
                     continue
             print_sep()
         fast_iso = True
@@ -521,7 +555,7 @@ def interactive_terminal():
 #                fast_iso = False
 #                break
 #            else:
-#                print('Wrong input')
+#                input("\nWrong Input. Press Enter to try again.\n")
 #                continue
         high_res = False
 #        if not fast_iso:
@@ -534,7 +568,7 @@ def interactive_terminal():
 #                if var == 'n':
 #                    break
 #                else:
-#                    print('Wrong input')
+#                    input("\nWrong Input. Press Enter to try again.\n")
 #                    continue
 #            if input_order[0] != 1:
 #                print("")
@@ -571,7 +605,7 @@ def interactive_terminal():
                 if var == 'n':
                     break
                 else:
-                    print('Wrong input')
+                    input("\nWrong Input. Press Enter to try again.\n")
                     continue
             print("")
             if ms2[0]:
@@ -583,7 +617,7 @@ def interactive_terminal():
                     if var == 'n':
                         break
                     else:
-                        print('Wrong input')
+                        input("\nWrong Input. Press Enter to try again.\n")
                         continue
                 print("")
             accuracy_unit = "mz"
@@ -595,7 +629,7 @@ def interactive_terminal():
                 if var == 'mz' or var == '':
                     break
                 else:
-                    print('Wrong input')
+                    input("\nWrong Input. Press Enter to try again.\n")
                     continue
             print("")
             accuracy_value = 0.0
@@ -604,7 +638,7 @@ def interactive_terminal():
                 try:
                     var = float(var)
                 except:
-                    print('Wrong input')
+                    input("\nWrong Input. Press Enter to try again.\n")
                     continue
                 accuracy_value = var
                 break
@@ -617,7 +651,7 @@ def interactive_terminal():
                 try:
                     var = float(var)
                 except:
-                    print('Wrong input')
+                    input("\nWrong Input. Press Enter to try again.\n")
                     continue
                 rt_int[0] = var
                 break
@@ -629,7 +663,7 @@ def interactive_terminal():
                 try:
                     var = float(var)
                 except:
-                    print('Wrong input')
+                    input("\nWrong Input. Press Enter to try again.\n")
                     continue
                 rt_int[1] = var
                 break
@@ -643,7 +677,7 @@ def interactive_terminal():
                 try:
                     var = int(var)
                 except:
-                    print('Wrong input')
+                    input("\nWrong Input. Press Enter to try again.\n")
                     continue
                 max_ppm = var
                 break
@@ -656,10 +690,10 @@ def interactive_terminal():
                 try:
                     var = float(var)
                 except:
-                    print('Wrong input')
+                    input("\nWrong Input. Press Enter to try again.\n")
                     continue
                 if var < 0.0 or var > 1.0:
-                    print('Wrong input')
+                    input("\nWrong Input. Press Enter to try again.\n")
                     continue
                 iso_fit = var
                 break
@@ -672,10 +706,10 @@ def interactive_terminal():
                 try:
                     var = float(var)
                 except:
-                    print('Wrong input')
+                    input("\nWrong Input. Press Enter to try again.\n")
                     continue
                 if var < 0.0 or var > 1.0:
-                    print('Wrong input')
+                    input("\nWrong Input. Press Enter to try again.\n")
                     continue
                 curve_fit = var
                 break
@@ -688,7 +722,7 @@ def interactive_terminal():
                 try:
                     var = int(var)
                 except:
-                    print('Wrong input')
+                    input("\nWrong Input. Press Enter to try again.\n")
                     continue
                 sn = var
                 break
@@ -757,7 +791,7 @@ def interactive_terminal():
             try:
                 var = int(var)
             except:
-                print('Wrong input')
+                input("\nWrong Input. Press Enter to try again.\n")
                 continue
             max_ppm = var
             break
@@ -768,10 +802,10 @@ def interactive_terminal():
             try:
                 var = float(var)
             except:
-                print('Wrong input')
+                input("\nWrong Input. Press Enter to try again.\n")
                 continue
             if var < 0.0 or var > 1.0:
-                print('Wrong input')
+                input("\nWrong Input. Press Enter to try again.\n")
                 continue
             iso_fit = var
             break
@@ -782,10 +816,10 @@ def interactive_terminal():
             try:
                 var = float(var)
             except:
-                print('Wrong input')
+                input("\nWrong Input. Press Enter to try again.\n")
                 continue
             if var < 0.0 or var > 1.0:
-                print('Wrong input')
+                input("\nWrong Input. Press Enter to try again.\n")
                 continue
             curve_fit = var
             break
@@ -796,7 +830,7 @@ def interactive_terminal():
             try:
                 var = int(var)
             except:
-                print('Wrong input')
+                input("\nWrong Input. Press Enter to try again.\n")
                 continue
             sn = var
             break
@@ -812,7 +846,7 @@ def interactive_terminal():
             if var == 'n':
                 break
             else:
-                print('Wrong input')
+                input("\nWrong Input. Press Enter to try again.\n")
                 continue
         path = default_path
         while True:
@@ -832,4 +866,212 @@ def interactive_terminal():
                 path = var
                 break
         return input_order, commented, path
+        
+def CLI():
+    '''A function that handles the input of parameters through the CLI 
+    and turns them into arguments for the functions used in core module.
+    
+    Parameters
+    ----------
+    none
+    
+    Uses
+    ----
+    pathlib.Path
+        Class of paths that can be used to create, change and access 
+        directories in the computer.
+    
+    Returns
+    -------
+    functions arguments
+        Organized arguments for the various functions used in core module.
+    '''
+    custom_glycans_list = [False, []]
+    min_max_monos = [0, 0]
+    min_max_hex = [0, 0]
+    min_max_hexnac = [0, 0]
+    min_max_xyl = [0, 0]
+    min_max_sia = [0, 0]
+    min_max_fuc = [0, 0]
+    min_max_ac = [0, 0]
+    min_max_gc = [0, 0]
+    force_nglycan = False
+    max_adducts = {}
+    adducts_exclusion = []
+    max_charges = 0
+    reducing_end_tag = 0.0
+    permethylated = False
+    reduced = False
+    lactonized_ethyl_esterified = False
+    fast_iso = True
+    high_res = False
+    internal_standard = 0.0
+    imp_exp_library = [False, False]
+    exp_lib_name = ''
+    library_path = ''
+    only_gen_lib = False
+
+    multithreaded_analysis = True
+    number_cores = 'all'
+    analyze_ms2 = [False, False, False]
+    reporter_ions = []
+    tolerance = ['mz', 0.01]
+    ret_time_interval = [0, 99999, 0.2]
+    rt_tolerance_frag = 0.2
+    min_isotopologue_peaks = 2
+    min_ppp = [False, 0]
+    close_peaks = [False, 3]
+    align_chromatograms = True
+    percentage_auc = 0.1
+    max_ppm = 10
+    iso_fit_score = 0.9
+    curve_fit_score = 0.9
+    s_to_n = 3
+    custom_noise = [False, []]
+    samples_path = ''
+    save_path = ''
+    plot_metaboanalyst = [False, []]
+    compositions = True
+    iso_fittings = False
+    reanalysis = False
+    reanalysis_path = ''
+    output_plot_data = False
+
+    samples_list = []
+    samples_names = []
+    
+    parameters = interactive_terminal()
+    Execution_Functions.print_sep()
+    if parameters[0][0] == 1 or parameters[0][0] == 2:
+        if parameters[0][1] == 1:
+            custom_glycans_list = (True, parameters[1])
+            max_adducts = parameters[2]
+            max_charges = parameters[3]
+            reducing_end_tag = parameters[4]
+            fast_iso = parameters[5]
+            high_res = parameters[6]
+            permethylated = parameters[8]
+            reduced = parameters[9]
+        if parameters[0][1] == 2:
+            min_max_monos = (parameters[1][0], parameters[1][1])
+            min_max_hex = (parameters[1][2], parameters[1][3])
+            min_max_hexnac = (parameters[1][4], parameters[1][5])
+            min_max_xyl = (parameters[1][15], parameters[1][16])
+            min_max_fuc = (parameters[1][6], parameters[1][7])
+            min_max_ac = (parameters[1][8], parameters[1][9])
+            min_max_gc = (parameters[1][10], parameters[1][11])
+            min_max_sia = (parameters[1][12], parameters[1][13])
+            force_nglycan = parameters[1][14]
+            max_adducts = parameters[2]
+            max_charges = parameters[3]
+            reducing_end_tag = parameters[4]
+            fast_iso = parameters[5]
+            high_res = parameters[6]
+            permethylated = parameters[8]
+            reduced = parameters[9]
+        if parameters[0][0] == 1:
+            lactonized_ethyl_esterified = parameters[10]
+            if parameters[0][1] == 1:
+                force_nglycan = parameters[11]
+            save_path = parameters[7]
+            if save_path[-1] != "/":
+                save_path+= "/"
+            only_gen_lib = True
+        if parameters[0][0] == 2:
+            analyze_ms2 = parameters[7]
+            accuracy_unit = parameters[8]
+            accuracy_value = parameters[9]
+            tolerance = [accuracy_unit, accuracy_value]
+            ret_time_interval = (parameters[10][0], parameters[10][1], 0.2)
+            min_isotopologue_peaks = parameters[11]
+            max_ppm = parameters[12]
+            iso_fit_score = parameters[13]
+            curve_fit_score = parameters[14]
+            s_to_n = parameters[15]
+            samples_path = parameters[16]
+            samples_list = Execution_Functions.samples_path_to_list(samples_path)
+            if len(samples_list) == 0:                                             
+                if not os.isatty(0):
+                    Execution_Functions.print_sep()
+                    print("No sample files to analyze.")
+                    os._exit(1)
+                else:
+                    Execution_Functions.print_sep()
+                    input("No sample files to analyzed. Press Enter\nto exit.")
+                    os._exit(1)
+            
+            samples_names = Execution_Functions.sample_names(samples_list)
+            print("Sample files detected: "+str(len(samples_names)))
+            for i in samples_names:
+                print("--> "+i)
+            Execution_Functions.print_sep()
+            save_path = parameters[17]
+            permethylated = parameters[18]
+            reduced = parameters[19]
+            lactonized_ethyl_esterified = parameters[20]
+            if parameters[0][1] == 1:
+                force_nglycan = parameters[21]
+    if parameters[0][0] == 3:
+        save_path = parameters[1]
+        max_ppm = parameters[2]
+        iso_fit_score = parameters[3]
+        curve_fit_score = parameters[4]
+        s_to_n = parameters[5]
+        reanalysis = (True, True)
+    if parameters[0][0] == 4:
+        comments = parameters[1]
+        save_path = parameters[2]
+        Execution_Functions.generate_cfg_file(save_path, comments)
+        
+    if save_path != '':
+        pathlib.Path(save_path).mkdir(exist_ok = True, parents = True)
+        
+    if library_path != '':
+        shutil.copy(library_path, os.path.join(temp_folder, 'glycans_library.py'))
+        spec = importlib.util.spec_from_file_location("glycans_library", temp_folder+"/glycans_library.py")
+        lib_module = importlib.util.module_from_spec(spec)
+        spec.loader.exec_module(lib_module)
+        try:
+            library_metadata = lib_module.metadata
+        except:
+            library_metadata = []
+        if len(library_metadata) > 0:
+            min_max_monos = library_metadata[0]
+            min_max_hex = library_metadata[1]
+            min_max_hexnac = library_metadata[2]
+            min_max_fuc = library_metadata[3]
+            min_max_sia = library_metadata[4]
+            min_max_ac = library_metadata[5]
+            min_max_gc = library_metadata[6]
+            force_nglycan = library_metadata[7]
+            max_adducts = library_metadata[8]
+            max_charges = library_metadata[9]
+            tag_mass = library_metadata[10]
+            internal_standard = library_metadata[11]
+            permethylated = library_metadata[12]
+            lactonized_ethyl_esterified = library_metadata[13]
+            reduced = library_metadata[14]
+            fast_iso = library_metadata[15]
+            high_res = library_metadata[16]   
+            if len(library_metadata) > 18:
+                min_max_xyl = library_metadata[18] 
+        
+    #args to execution functions:
+    output_filtered_data_args = [curve_fit_score, iso_fit_score, s_to_n, max_ppm, percentage_auc, reanalysis, reanalysis_path, save_path, analyze_ms2[0], analyze_ms2[2], reporter_ions, plot_metaboanalyst, compositions, align_chromatograms, force_nglycan, ret_time_interval[2], rt_tolerance_frag, iso_fittings, output_plot_data, multithreaded_analysis, number_cores, 0.0]
+
+    imp_exp_gen_library_args = [custom_glycans_list, min_max_monos, min_max_hex, min_max_hexnac, min_max_xyl, min_max_sia, min_max_fuc, min_max_ac, min_max_gc, force_nglycan, max_adducts, adducts_exclusion, max_charges, reducing_end_tag, fast_iso, high_res, imp_exp_library, library_path, exp_lib_name, only_gen_lib, save_path, internal_standard, permethylated, lactonized_ethyl_esterified, reduced]
+
+    list_of_data_args = [samples_list]
+
+    index_spectra_from_file_ms1_args = [None, 1, multithreaded_analysis, number_cores]
+
+    index_spectra_from_file_ms2_args = [None, 2, multithreaded_analysis, number_cores]
+
+    analyze_files_args = [None, None, None, None, tolerance, ret_time_interval, min_isotopologue_peaks, min_ppp, max_charges, custom_noise, close_peaks, multithreaded_analysis, number_cores]
+
+    analyze_ms2_args = [None, None, None, ret_time_interval, tolerance, min_max_monos, min_max_hex, min_max_hexnac, min_max_xyl,  min_max_sia, min_max_fuc, min_max_ac, min_max_gc, max_charges, reducing_end_tag, force_nglycan, permethylated, reduced, lactonized_ethyl_esterified, analyze_ms2[1], analyze_ms2[2], ret_time_interval[2], multithreaded_analysis, number_cores]
+
+    arrange_raw_data_args = [None, samples_names, analyze_ms2[0], save_path, [(custom_glycans_list, min_max_monos, min_max_hex, min_max_hexnac, min_max_sia, min_max_fuc, min_max_ac, min_max_gc, force_nglycan, max_adducts, adducts_exclusion, max_charges, reducing_end_tag, permethylated, reduced, lactonized_ethyl_esterified, fast_iso, high_res, internal_standard, imp_exp_library, exp_lib_name, library_path, only_gen_lib, min_max_xyl), (multithreaded_analysis, number_cores, analyze_ms2, reporter_ions, tolerance, ret_time_interval, rt_tolerance_frag, min_isotopologue_peaks, min_ppp, close_peaks, align_chromatograms, percentage_auc, max_ppm, iso_fit_score, curve_fit_score, s_to_n, custom_noise, samples_path, save_path, plot_metaboanalyst, compositions, iso_fittings, reanalysis, reanalysis_path, output_plot_data)]]
+
+    return output_filtered_data_args, imp_exp_gen_library_args, list_of_data_args, index_spectra_from_file_ms1_args, index_spectra_from_file_ms2_args, analyze_files_args, analyze_ms2_args, arrange_raw_data_args, samples_names, reanalysis, analyze_ms2[0]
     
\ No newline at end of file
