@@ -28,6 +28,10 @@ from scipy import sparse
 import numpy
 import sys
 import datetime
+import xlsxwriter
+import zipfile
+import pathlib
+import os
 
 ##---------------------------------------------------------------------------------------
 ##Hard-coded permanent information
@@ -35,9 +39,10 @@ import datetime
 monosaccharides = {
     "H": ("Hexose", "C6O6H12", {"C": 6, "O": 5, "N": 0, "H": 10}),
     "N": ("N-Acetyl Hexosamine", "C8O6NH15", {"C": 8, "O": 5, "N": 1, "H": 13}),
+    "X": ("Xylose", "C5O5H10", {"C": 5, "O": 4, "N": 0, "H": 8}),
     "S": ("Acetyl Neuraminic Acid", "C11O9NH19", {"C": 11, "O": 8, "N": 1, "H": 17}),
-    "lS": ("Lactonized Acetyl Neuraminic Acid alpha2,3 bound", "C11O8NH17", {"C": 11, "O": 7, "N": 1, "H": 15}),
-    "eS": ("Ethyl-Esterified Acetyl Neuraminic Acid alpha2,6 bound", "C13O9NH23", {"C": 13, "O": 8, "N": 1, "H": 21}),
+    "Am": ("Lactonized Acetyl Neuraminic Acid alpha2,3 bound", "C11O8N2H20", {"C": 11, "O": 7, "N": 2, "H": 18}),
+    "E": ("Ethyl-Esterified Acetyl Neuraminic Acid alpha2,6 bound", "C13O9NH23", {"C": 13, "O": 8, "N": 1, "H": 21}),
     "F": ("Fucose", "C6O5H12", {"C": 6, "O": 4, "N": 0, "H": 10}),
     "G": ("Glycolyl Neuraminic Acid", "C11O10NH19", {"C": 11, "O": 9, "N": 1, "H": 17})
     }
@@ -55,6 +60,40 @@ times during a run.
 ##General functions (these functions use only external libraries, such as itertools and
 ##pyteomics).
 
+def binary_search_with_tolerance(arr, target, low, high, tolerance):
+    '''
+    '''
+    # Base case: if the range is invalid, the target is not in the array
+    if low > high:
+        return -1  # Target not found
+    
+    # Find the middle index
+    mid = (low + high) // 2
+    
+    # Check if the target is within the tolerance range of the middle element
+    if abs(arr[mid] - target) <= tolerance:
+        mid_distance = abs(arr[mid] - target)
+        smallest_distance = mid_distance
+        smallest_distance_id = mid
+        if mid != len(arr)-1 and arr[mid+1] <= tolerance:
+            mid_plus_distance = abs(arr[mid+1] - target)
+            if mid_plus_distance < smallest_distance:
+                smallest_distance = mid_plus_distance
+                smallest_distance_id = mid+1
+        if mid != 0 and arr[mid-1] <= tolerance:
+            mid_minus_distance = abs(arr[mid-1] - target)
+            if mid_minus_distance < smallest_distance:
+                smallest_distance = mid_minus_distance
+                smallest_distance_id = mid-1
+                
+        return smallest_distance_id  # Closest target found within tolerance at or around index mid
+    elif arr[mid] < target:
+        # If target is greater, ignore the left half
+        return binary_search_with_tolerance(arr, target, mid + 1, high, tolerance)
+    else:
+        # If target is smaller, ignore the right half
+        return binary_search_with_tolerance(arr, target, low, mid - 1, tolerance)
+
 def linear_regression(x, y, th = 2.5):
     '''Traces a linear regression of supplied 2d data points and returns the slope,
     y-intercept and the indices of the outliers outside the determined threshold.
@@ -106,6 +145,84 @@ def linear_regression(x, y, th = 2.5):
     outlier_indices = where(abs(residuals) > threshold)[0]
     # Final equation is used as (y = mx + b)
     return m, b, outlier_indices
+    
+def cleanup_by_average(x, y, th = 2.5):
+    '''
+    '''
+    if len(x) != len(y): # Ensure x and y have the same length
+        raise ValueError("Input arrays x and y must have the same length.")
+    if len(x) == 0:
+        return 0, 0, []
+    if len(x) == 1:
+        return 0, y[0], []
+    x = array(x) # Convert input data to numpy arrays
+    y = array(y) # Convert input data to numpy arrays
+    predicted_y = mean(y)
+    residuals = y - predicted_y
+    threshold = th * std(residuals)
+    outlier_indices = where(abs(residuals) > threshold)[0]
+    # Final equation is used as (y = mx + b)
+    return mean, residuals, outlier_indices
+    
+def make_gg(temp_dir, save_dir, filename):
+    '''Creates a zipped file with extension .gg containing raw_data files.
+    
+    Parameters
+    ----------
+    temp_dir : string
+        A string containing the path to the temp_dir where raw_data files are.
+        
+    save_dir : string
+        A string containing the path to save the .gg file.
+        
+    filename : string
+        The name of the .gg file.
+        
+    Uses
+    ----
+    zipfile.ZipFile
+        Creates a zipfile object to add files to.
+    
+    zipfile.write
+        Adds files to the zipfile object.
+        
+    Returns
+    -------
+    nothing
+        Creates filename.gg file, containing files in temp_dir, at the save_dir.
+    '''
+    files_list = os.listdir(temp_dir)
+    with zipfile.ZipFile(save_dir+filename+".gg", 'w') as zipf:
+        for file in files_list:
+            zipf.write(temp_dir+"/"+str(file), arcname=file)
+            
+def open_gg(gg_file, temp_path):
+    '''Unzipts the .gg file to temp_path.
+    
+    Parameters
+    ----------
+    gg_file : string
+        Path to the .gg file.
+        
+    temp_path : string
+        String containing the path to the folder where the files will be extracted.
+        
+    Uses
+    ----
+    zipfile.ZipFile
+        Creates a zipfile object to add files to.
+        
+    zipfile.extractall
+        Extract all files in .gg file to the target directory.
+        
+    Returns
+    -------
+    nothing
+        Extracts raw_data files from gg_file to temp_path.
+    '''
+    with zipfile.ZipFile(gg_file, 'r') as zip_ref:
+        zip_ref.extractall(temp_path)
+    
 
 def calculate_ppm_diff(mz, target):
     '''Calculates the PPM difference between a mz and a target mz.
@@ -155,6 +272,35 @@ def tolerance_calc(unit,
     else:
         return("Unit for tolerance not 'ppm' or 'mz'.")
         
+def autofit_columns_excel(df, worksheet):
+    '''Autofits the column width in a excel worksheet based on a dataframe used to make it.
+    
+    Parameters
+    ----------
+    df : Pandas Dataframe
+        Dataframe containing the data used to make the worksheet with Pandas.
+        
+    worksheet : Worksheet XLSXwriter object
+        The worksheet to autofit the columns.
+        
+    Uses
+    ----
+    xlsxwriter
+        Library to write and edit excel files and objects.
+        
+    Returns
+    -------
+    nothing
+        Directly edits the worksheet.
+    '''
+    for idx, col in enumerate(df):  # loop through all columns
+        series = df[col]
+        max_len = max((
+            series.astype(str).map(len).max(),  # len of largest item
+            len(str(series.name))  # len of column name/header
+            )) + 1  # adding a little extra space
+        worksheet.set_column(idx, idx, max_len)
+        
 def speyediff(N, d, format='csc'):
     '''Construct a d-th order sparse difference matrix based on an initial 
     N x N identity matrix. Obtained from https://github.com/mhvwerts/whittaker-eilers-smoother, 
@@ -238,6 +384,9 @@ def rt_noise_level_parameters_set(mz_int, mode):
     
     noise = []
     for j_j, j in enumerate(segments_list):
+        if len(j) == 0:
+            noise.append(1.0)
+            continue
         intensity_std = numpy.std(j)
         noise_threshold = 2.0 * intensity_std
         if (min(j) != 0 and noise_threshold > min(j)*5) or noise_threshold > max(j)*0.5: #this means that the data is denoised already, so it picks really high intensity as possible noise
@@ -443,10 +592,8 @@ def count_seq_letters(string):
             current_letter = i
             count = string.count(i)
             if i == "L":
-                friendly_letters['lS'] = count
-            if i == "E":
-                friendly_letters['eS'] = count
-            elif i != "L" and i != "E":
+                friendly_letters['Am'] = count
+            elif i != "L":
                 friendly_letters[i] = count
     return friendly_letters
 
@@ -488,7 +635,7 @@ def sum_monos(*compositions):
     summed_comp : dict
         Dictionary containing the sum of each monosaccharides of the compositions.
     '''
-    summed_comp = {"H": 0, "N": 0, "S": 0, "lS": 0, "eS": 0, "F": 0, "G": 0, "T": 0}
+    summed_comp = {"H": 0, "N": 0, "X": 0, "S": 0, "Am": 0, "E": 0, "F": 0, "G": 0, "T": 0}
     for i in compositions:
         for j in i:
             summed_comp[j]+=i[j]
@@ -629,8 +776,8 @@ def calculate_isotopic_pattern(glycan_atoms,
     return relative_isotop_pattern, relative_isotop_mass
 
 def gen_adducts_combo(adducts,
-                      exclusions,
-                      max_charge):
+                      exclusions=[],
+                      max_charge=3):
     '''Generates a list of dictionaries with compositions of adducts combinations, based
     on parameters set.
 
