@@ -16,13 +16,14 @@
 # or by typing 'license' after running it stand-alone in the terminal
 # by typing 'glycogenius'. If not, see <https://www.gnu.org/licenses/>.
 
-import pathlib
-import importlib
 from . import General_Functions
 from pyteomics import mzxml, mzml, mass, auxiliary
 from itertools import combinations_with_replacement
+from collections import Counter
 from re import split
 from math import inf
+import pathlib
+import importlib
 import sys
 import datetime
 
@@ -30,9 +31,45 @@ import datetime
 ##Library generating-associated functions (these functions make vast use of the general
 ## functions).
 
+def generate_combinations_with_constraints(characters, length, constraints):
+    '''
+    '''
+    def is_valid(counts):
+        for char, (min_count, max_count) in constraints.items():
+            count = counts[char]
+            if count < min_count or count > max_count:
+                return False
+        return True
+
+    # Generate all possible count combinations within the constraints
+    char_indices = list(range(len(characters)))
+    valid_count_combinations = []
+
+    for combination in combinations_with_replacement(char_indices, length):
+        counts = Counter(combination)
+        counts = {characters[idx]: counts[idx] for idx in counts}
+        
+        # Fill in missing characters with zero counts
+        for char in characters:
+            if char not in counts:
+                counts[char] = 0
+        
+        if is_valid(counts):
+            valid_count_combinations.append(counts)
+
+    # Convert count combinations to strings
+    valid_combinations = []
+    for counts in valid_count_combinations:
+        if 'L' in counts:
+            counts['Am'] = counts.pop('L')
+        valid_combinations.append(counts)
+
+    return valid_combinations
+    
 def generate_glycans_library(min_max_mono,
                              min_max_hex,
                              min_max_hexnac,
+                             min_max_xyl,
                              min_max_sialics,
                              min_max_fuc,
                              min_max_ac,
@@ -104,40 +141,50 @@ def generate_glycans_library(min_max_mono,
         glycans generated.
     '''
     glycans = []
-    def_glycan_comp = {"H": 0, "N": 0, "S": 0, "lS": 0, "eS": 0, "F": 0, "G": 0}
+    def_glycan_comp = {"H": 0, "N": 0, "X": 0, "S": 0, "Am": 0, "E": 0, "F": 0, "G": 0}
+    
+    if lactonized_ethyl_esterified:
+        monos_chars = "HNXLEFG"
+    else:
+        monos_chars = "HNXSFG"
+        
+    constraints = {'H': (min_max_hex[0], min_max_hex[1]),
+                   'N': (min_max_hexnac[0], min_max_hexnac[1]),
+                   'X': (min_max_xyl[0], min_max_xyl[1]),
+                   'F': (min_max_fuc[0], min_max_fuc[1]),
+                   'G': (min_max_gc[0], min_max_gc[1])}
+                   
+    if lactonized_ethyl_esterified:
+        constraints['L'] = (min_max_ac[0], min_max_ac[1])
+        constraints['E'] = (min_max_ac[0], min_max_ac[1])
+    else:
+        constraints['S'] = (min_max_ac[0], min_max_ac[1])
+        
     for i in range(min_max_mono[0], min_max_mono[1]+1):
-        if lactonized_ethyl_esterified:
-            for j in combinations_with_replacement("HNLEFG", i):
-                glycans.append(General_Functions.sum_monos(def_glycan_comp, General_Functions.count_seq_letters("".join(j))))
-        else:
-            for j in combinations_with_replacement("HNSFG", i):
-                glycans.append(General_Functions.sum_monos(def_glycan_comp, General_Functions.count_seq_letters("".join(j))))
+        combinations = generate_combinations_with_constraints(monos_chars, i, constraints)
+        for j in combinations:
+            glycans.append(General_Functions.sum_monos(def_glycan_comp, j))
+            
     to_be_removed = []
     for i_i, i in enumerate(glycans):
         if lactonized_ethyl_esterified:
-            if ((i['H'] < min_max_hex[0]) or (i['H'] > min_max_hex[1])
-                or (i['N'] < min_max_hexnac[0]) or (i['N'] > min_max_hexnac[1])
-                or (i['lS']+i['eS']+i['G'] < min_max_sialics[0])
-                or (i['lS']+i['eS']+i['G'] > min_max_sialics[1])
-                or (i['F'] < min_max_fuc[0]) or (i['F'] > min_max_fuc[1])
-                or (i['lS']+i['eS'] < min_max_ac[0]) or (i['lS']+i['eS'] > min_max_ac[1])
-                or (i['G'] < min_max_gc[0]) or (i['G'] > min_max_gc[1])):
+            if ((i['Am']+i['E']+i['G'] < min_max_sialics[0])
+                or (i['Am']+i['E']+i['G'] > min_max_sialics[1])
+                or (i['Am']+i['E'] < min_max_ac[0]) 
+                or (i['Am']+i['E'] > min_max_ac[1])):
                 to_be_removed.append(i)
         else:
-            if ((i['H'] < min_max_hex[0]) or (i['H'] > min_max_hex[1])
-                or (i['N'] < min_max_hexnac[0]) or (i['N'] > min_max_hexnac[1])
-                or (i['S']+i['G'] < min_max_sialics[0])
-                or (i['S']+i['G'] > min_max_sialics[1])
-                or (i['F'] < min_max_fuc[0]) or (i['F'] > min_max_fuc[1])
-                or (i['S'] < min_max_ac[0]) or (i['S'] > min_max_ac[1])
-                or (i['G'] < min_max_gc[0]) or (i['G'] > min_max_gc[1])):
+            if ((i['S']+i['G'] < min_max_sialics[0])
+                or (i['S']+i['G'] > min_max_sialics[1])):
                 to_be_removed.append(i)
     if n_glycan:
         if lactonized_ethyl_esterified:
             for i_i, i in enumerate(glycans):
-                if ((i['lS']+i['eS']+i['G'] > i['N']-2)
-                    or (i['F'] >= i['N']) or (i['lS']+i['eS']+i['G'] > i['H']-2)
-                    or (i['H'] < 3) or (i['N'] < 2)):
+                if ((i['Am']+i['E']+i['G'] > i['N']-2)
+                    or (i['F'] >= i['N']) 
+                    or (i['Am']+i['E']+i['G'] > i['H']-2)
+                    or (i['H'] < 3) 
+                    or (i['N'] < 2)):
                     if i not in to_be_removed:
                         to_be_removed.append(i)
         else:
@@ -254,6 +301,15 @@ def full_glycans_library(library,
         glycans with the desired adducts combination.
     '''
     full_library = {}
+    
+    try:
+        tag_mass = float(tag_mass)
+    except:
+        if tag_mass.split('-')[0] == 'pep':
+            tag_mass = dict(mass.Composition(sequence = tag_mass.split('-')[-1]))
+            tag_mass['H'] -= 2
+            tag_mass['O'] -= 1
+            
     if tag_mass != 0:
         if type(tag_mass) == float:
             tag = General_Functions.calculate_comp_from_mass(tag_mass)
@@ -337,7 +393,7 @@ def full_glycans_library(library,
         for i in range(len(i_iso_dist[0])):
             i_iso_dist[1].append(internal_standard+(i*General_Functions.h_mass))
         full_library[i_formula] = {}
-        full_library[i_formula]['Monos_Composition'] = {"H": 0, "N": 0, "S": 0, "lS": 0, "eS": 0, "F": 0, "G": 0}
+        full_library[i_formula]['Monos_Composition'] = {"H": 0, "N": 0, "X": 0, "S": 0, "Am": 0, "E": 0, "F": 0, "G": 0}
         full_library[i_formula]['Neutral_Mass'] = i_neutral_mass
         full_library[i_formula]['Neutral_Mass+Tag'] = i_neutral_mass
         full_library[i_formula]['Isotopic_Distribution'] = i_iso_dist[0]
@@ -352,6 +408,7 @@ def full_glycans_library(library,
 def fragments_library(min_max_mono,
                       min_max_hex,
                       min_max_hexnac,
+                      min_max_xyl,
                       min_max_sialics,
                       min_max_fuc,
                       min_max_ac,
@@ -467,47 +524,53 @@ def fragments_library(min_max_mono,
     '''
     print("Building fragments library...", end = "", flush = True)
     glycans = []
-    def_glycan_comp = {"H": 0, "N": 0, "S": 0, "lS": 0, "eS": 0, "F": 0, "G": 0, "T" : 0}
-    for i in range(1, 18):
-        if lactonized_ethyl_esterified:
-            for j in combinations_with_replacement("HNLEFG", i):
-                glycans.append(General_Functions.sum_monos(def_glycan_comp, General_Functions.count_seq_letters("".join(j))))
-        else:
-            for j in combinations_with_replacement("HNSFGT", i):
-                glycans.append(General_Functions.sum_monos(def_glycan_comp, General_Functions.count_seq_letters("".join(j))))
+    def_glycan_comp = {"H": 0, "N": 0, "X": 0, "S": 0, "Am": 0, "E": 0, "F": 0, "G": 0, "T" : 0}
+    
+    if lactonized_ethyl_esterified:
+        monos_chars = "HNXLEFGT"
+    else:
+        monos_chars = "HNXSFGT"
+        
+    constraints = {'H': (0, min_max_hex[1]),
+                   'N': (0, min_max_hexnac[1]),
+                   'X': (0, min_max_xyl[1]),
+                   'F': (0, min_max_fuc[1]),
+                   'G': (0, min_max_gc[1]),
+                   'T': (0, 1)}
+                   
+    if lactonized_ethyl_esterified:
+        constraints['L'] = (0, min_max_ac[1])
+        constraints['E'] = (0, min_max_ac[1])
+    else:
+        constraints['S'] = (0, min_max_ac[1])
+        
+    for i in range(1, min_max_mono[1]):
+        combinations = generate_combinations_with_constraints(monos_chars, i, constraints)
+        for j in combinations:
+            glycans.append(General_Functions.sum_monos(def_glycan_comp, j))
+            
     to_be_removed = []
     for i_i, i in enumerate(glycans):
         if lactonized_ethyl_esterified:
-            if ((i['T'] > 1) 
-                or (i['T'] == 1 and i['N'] == 0)
-                or (i['H'] > min_max_hex[1])
-                or (i['N'] > min_max_hexnac[1])
-                or (i['lS']+i['eS']+i['G'] > min_max_sialics[1])
-                or (i['F'] > min_max_fuc[1])
-                or (i['lS']+i['eS'] > min_max_ac[1])
-                or (i['G'] > min_max_gc[1])):
+            if ((i['T'] == 1 and i['N'] == 0)
+                or (i['Am']+i['E']+i['G'] > min_max_sialics[1])
+                or (i['Am']+i['E'] > min_max_ac[1])):
                 to_be_removed.append(i_i)
         else:
-            if ((i['T'] > 1) 
-                or (i['T'] == 1 and i['N'] == 0)
-                or (i['H'] > min_max_hex[1])
-                or (i['N'] > min_max_hexnac[1])
-                or (i['S']+i['G'] > min_max_sialics[1])
-                or (i['F'] > min_max_fuc[1])
-                or (i['S'] > min_max_ac[1])
-                or (i['G'] > min_max_gc[1])):
+            if ((i['T'] == 1 and i['N'] == 0)
+                or (i['S']+i['G'] > min_max_sialics[1])):
                 to_be_removed.append(i_i)
         if nglycan and i_i not in to_be_removed: #some rules and hardcoded exceptions for N-Glycans
             if lactonized_ethyl_esterified:
                 if ((i['T'] == 1 and sum(i.values()) < 8 and i['S']+i['G'] > 0)
                     or (sum(i.values()) < 6 and i['S'] >= 1 and i['N'] > 1)
                     or (i['H'] > 0 and i['T'] == 1 and i['N'] < 2)
-                    or (i['H'] == 2 and i['N'] == 1 and i['lS']+i['eS'] == 0 and i['F'] == 0 and i['G'] == 0 and i['T'] == 1)
-                    or (i['H'] == 0 and i['N'] == 1 and (i['lS']+i['eS'] == 1 or i['G'] == 1) and i['F'] == 0 and i['T'] == 0)
-                    or (i['H'] == 1 and i['N'] == 3 and i['lS']+i['eS'] == 0 and i['F'] == 0 and i['G'] == 0 and i['T'] == 1)
-                    or (i['H'] > 1 and i['N'] == 0 and (i['lS']+i['eS'] == 1 or i['G'] == 1) and i['F'] == 0 and i['T'] == 0)
-                    or (i['H'] == 1 and i['N'] == 1 and i['lS']+i['eS'] == 0 and i['F'] == 0 and i['G'] == 0 and i['T'] == 1)
-                    or (i['H'] == 3 and i['N'] == 1 and i['lS']+i['eS'] == 0 and i['F'] == 0 and i['G'] == 0 and i['T'] == 1)):
+                    or (i['H'] == 2 and i['N'] == 1 and i['Am']+i['E'] == 0 and i['F'] == 0 and i['G'] == 0 and i['T'] == 1)
+                    or (i['H'] == 0 and i['N'] == 1 and (i['Am']+i['E'] == 1 or i['G'] == 1) and i['F'] == 0 and i['T'] == 0)
+                    or (i['H'] == 1 and i['N'] == 3 and i['Am']+i['E'] == 0 and i['F'] == 0 and i['G'] == 0 and i['T'] == 1)
+                    or (i['H'] > 1 and i['N'] == 0 and (i['Am']+i['E'] == 1 or i['G'] == 1) and i['F'] == 0 and i['T'] == 0)
+                    or (i['H'] == 1 and i['N'] == 1 and i['Am']+i['E'] == 0 and i['F'] == 0 and i['G'] == 0 and i['T'] == 1)
+                    or (i['H'] == 3 and i['N'] == 1 and i['Am']+i['E'] == 0 and i['F'] == 0 and i['G'] == 0 and i['T'] == 1)):
                     to_be_removed.append(i_i)
             else:
                 if ((i['T'] == 1 and sum(i.values()) < 8 and i['S']+i['G'] > 0)
@@ -522,13 +585,24 @@ def fragments_library(min_max_mono,
                     to_be_removed.append(i_i)
     for i in sorted(to_be_removed, reverse = True):
         del glycans[i]
+    try:
+        tag_mass = float(tag_mass)
+    except:
+        if tag_mass.split('-')[0] == 'pep':
+            tag_mass = dict(mass.Composition(sequence = tag_mass.split('-')[-1]))
+            tag_mass['H'] -= 2
+            tag_mass['O'] -= 1
+            
     if tag_mass != 0:
         if type(tag_mass) == float:
             tag = General_Functions.calculate_comp_from_mass(tag_mass)
-        else:
+        elif type(tag_mass) != dict:
             comp_tag = General_Functions.form_to_comp(tag_mass)
             tag = (comp_tag, mass.calculate_mass(comp_tag))
-            tag_mass = tag[1]
+        else:
+            comp_tag = tag_mass
+            tag = (comp_tag, mass.calculate_mass(comp_tag))
+        tag_mass = tag[1]
     else:
         tag = ({"C": 0, "O": 0, "N": 0, "H": 0}, 0.0)
     adducts_combo = General_Functions.gen_adducts_combo({'H' : 2}, [], max_charges)
